/****************************************************************************
 *
 * CUDA_EGS.cu, Version 1.0.0 Mon 09 Jan 2012
 *
 * ----------------------------------------------------------------------------
 *
 * CUDA EGS
 * Copyright (C) 2012 CancerCare Manitoba
 *
 * The latest version of CUDA EGS and additional information are available online at 
 * http://www.physics.umanitoba.ca/~elbakri/cuda_egs/ and http://www.lippuner.ca/cuda_egs
 *
 * CUDA EGS is free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.                                       
 *                                                                           
 * CUDA EGS is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                              
 *                                                                           
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * ----------------------------------------------------------------------------
 *
 *   Contact:
 *
 *   Jonas Lippuner
 *   Email: jonas@lippuner.ca 
 *
 ****************************************************************************/

#define CUDA_EGS

#include "CUDA_EGS.h"
#include "output.c"
#include "media.c"
#include "init.cu"
#include "kernels.cu"


uint read_step_counts(ulong *this_total, ulong *grand_total) {
    clock_t start = clock();
    cudaMemcpy(h_total_step_counts, d_total_step_counts, sizeof(total_step_counts_t), cudaMemcpyDeviceToHost); ce(12001);
    
    for (uchar i = 0; i < NUM_CAT; i++) {
        this_total[i] = 0;
        for (uchar j = 0; j < SIMULATION_NUM_BLOCKS; j++)
            this_total[i] += (*h_total_step_counts)[j][i];
            
        grand_total[i] += this_total[i];
    }

    clock_t stop = clock();

    return stop - start;
}

int main(int argc, char **argv) {
    
    // record start time
    clock_t tic = clock();

	// check whether -h or --help was specified
	for (int i = 0; i < argc; i++) {
		if ((string(argv[i]).compare("-h") == 0) || (string(argv[i]).compare("--help") == 0)) {
			printf("\nCUDA EGS version 1.0.0\nUsage: CUDA_EGS -i input_file [-o output_prefix] [-s RNG_seed] [-g GPU_id]\n");
			exit(0);
		}
	}
	
	// get input file
	int idxI = -1;
	for (int i = 0; i < argc; i++) {
		if (string(argv[i]).compare("-i") == 0) {
			idxI = i;
			break;
		}
	}

	if ((idxI == -1) || (idxI == argc - 1)) {
		printf("ERROR (1001): No input file was specified. Please use the command line option \"-i input_file\" to specify the input file.\n");
		exit(1001);
	}

	input_file = argv[idxI + 1];
	FILE *input_f;
	if (fopen_s(&input_f, input_file, "r")) {
		printf("ERROR (1002): The input file \"%s\" could not be opened.\n", input_file);
		exit(1002);
	}
	fclose(input_f);

    // get the full path of the input file, otherwise GetPrivateProfileString etc. might not 
    // be able to read it
    GetFullPathName(input_file, CHARLEN, charBuffer, NULL);
    string input_file_input = string(charBuffer);
    input_file = input_file_input.c_str();


	// get output prefix
	bool outputPrefixGiven = false;
	string rootName = "";
	int idxO = -1;
	for (int i = 0; i < argc; i++) {
		if (string(argv[i]).compare("-o") == 0) {
			idxO = i;
			break;
		}
	}
	if (idxO != -1) {
		if (idxO == argc - 1) {
			printf("ERROR (1003): The command line option -o was given, but no output_prefix was specified. Please use the command line option \"-o output_prefix\" to specify the output_prefix.\n");
			exit(1003);
		}

		rootName = string(argv[idxO + 1]);
		checkOutputPrefix(1004, rootName);
		outputPrefixGiven = true;
	}

	// get RNG seed
	bool RNGSeedGiven = false;
	uint seed;
	int idxS = -1;
	for (int i = 0; i < argc; i++) {
		if (string(argv[i]).compare("-s") == 0) {
			idxS = i;
			break;
		}
	}
	if (idxS != -1) {
		if (idxS == argc - 1) {
			printf("ERROR (1005): The command line option -s was given, but no RNG_seed was specified. Please use the command line option \"-s RNG_seed\" to specify the RNG seed.\n");
			exit(1005);
		}
		uint givenSeed;
		if (sscanf(argv[idxS + 1], "%u", &givenSeed) != 1) {
			printf("ERROR (1006): The given RNG seed \"%s\" could not be parsed as an unsigned 4-byte integer.\n", argv[idxS + 1]);
			exit(1006);
		}
		
		seed = givenSeed;
		RNGSeedGiven = true;
	}

	// get GPU id
	bool GPUIdGiven = false;
	int GPUId = -1;
	int idxG = -1;
	for (int i = 0; i < argc; i++) {
		if (string(argv[i]).compare("-g") == 0) {
			idxG = i;
			break;
		}
	}
	if (idxG != -1) {
		if (idxG == argc - 1) {
			printf("ERROR (1007): The command line option -g was given, but no GPU_id was specified. Please use the command line option \"-g GPU_id\" to specify the GPU id.\n");
			exit(1007);
		}
		uint givenId;
		if (sscanf(argv[idxG + 1], "%d", &givenId) != 1) {
			printf("ERROR (1008): The given GPU id \"%s\" could not be parsed as a 4-byte integer.\n", argv[idxG + 1]);
			exit(1008);
		}
		
		GPUId = givenId;
		GPUIdGiven = true;
	}

	// first we need to get the output prefix so that we can create the log file
	bool defaultOutputPrefixUsed = false;
	if (!outputPrefixGiven) {
		GetPrivateProfileString("General", "output prefix", "<<missing>>", charBuffer, CHARLEN, input_file);
		if (string(charBuffer) == "<<missing>>") {
			defaultOutputPrefixUsed = true;
			rootName = "CUDA_EGS_";
		}
		else {
			rootName = string(charBuffer);
			checkOutputPrefix(2001, rootName);
		}
	}

    // get the absolute path of the output prefix
    GetFullPathName(rootName.c_str(), CHARLEN, charBuffer, NULL);
    rootName = string(charBuffer);

	// create log file
	fopen_s(&logfile, (rootName + "log.txt").c_str(), "w");
	
	// log time
	getTime();
	logstr("\nCUDA EGS version 1.0.0 started at %s\n", charBuffer);

	// log info
    logstr("\nSimulation\n");
	logstr("  Input file  . . . . . . . . %s\n", input_file);
	if (defaultOutputPrefixUsed)
		logstr("WARNING: No output prefix was specified. Default value is used.\n");
	logstr("  Output prefix . . . . . . . %s\n", rootName.c_str());

    // get GPU
    bool defaultGPUUsed = false;
    if (!GPUIdGiven) {
        // first check whether a GPU id is given
        GetPrivateProfileString("General", "GPU id", "<<missing>>", charBuffer, CHARLEN, input_file);
		if (string(charBuffer) == "<<missing>>") {
			// no GPU id is given, check whether a GPU name is given
            GetPrivateProfileString("General", "GPU", "<<missing>>", charBuffer, CHARLEN, input_file);
            if (string(charBuffer) == "<<missing>>") {
                defaultGPUUsed = true;
			    GPUId = 0;
		    }
            else {
                int deviceCount = 0;
                cudaGetDeviceCount(&deviceCount); ce(2002);
                
                // find first device with matching name
                for (int dev = 0; dev < deviceCount; dev++) {
                    cudaDeviceProp deviceProp;
                    cudaGetDeviceProperties(&deviceProp, dev); ce(2022);
                    if (string(deviceProp.name).compare(string(charBuffer)) == 0) {
                        GPUId = dev;
                        break;
                    }
                }

                if (GPUId == -1)
                    error(2003, "No GPU with name \"%s\" was found.\n", charBuffer);
            }
        }
		else {
			int givenId;
		    if (sscanf(charBuffer, "%d", &givenId) != 1)
			    error(2004, "The given GPU id \"%s\" could not be parsed as a 4-byte integer.\n", charBuffer);
            GPUId = givenId;
		}
    }

    // check whether GPU has CUDA capability 2.0
    int deviceCount = 0;
    cudaGetDeviceCount(&deviceCount); ce(2005);
    if (deviceCount <= 0)
        error(2006, "No CUDA capabale GPUs found.\n");
    if (GPUId > deviceCount - 1)
        error(2007, "No CUDA capable GPU with id %d is available.\n", GPUId);
    
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, GPUId); ce(2021);
    if (defaultGPUUsed)
		logstr("WARNING: No GPU was specified. Default GPU is used.\n");
    logstr("  GPU . . . . . . . . . . . . %s (id %d)\n", deviceProp.name, GPUId);

    if (deviceProp.major < 2)
        error(2008, "CUDA capability 2.0 or higher is required, but the specified GPU only supports CUDA capability %d.%d.\n", deviceProp.major, deviceProp.minor);
    if (deviceProp.warpSize != WARP_SIZE)
        error(2009, "The code was compiled for a warp size of %d, but the GPU uses a warp size of %d. Please change the value of the constant WARP_SIZE in CUDA_EGS.h and recompile.\n", WARP_SIZE, deviceProp.warpSize);
    if (deviceProp.multiProcessorCount != NUM_MULTIPROC)
        error(2010, "The code was compiled for %d multiprocessors, but the GPU has %d multiprocessors. Please change the value of the constant NUM_MULTIPROC in CUDA_EGS.h and recompile.\n", NUM_MULTIPROC, deviceProp.multiProcessorCount);

    logstr("    Warp size . . . . . . . . %d\n", deviceProp.warpSize);
    logstr("    Multiprocessors . . . . . %d\n", deviceProp.multiProcessorCount);

    cudaSetDevice(GPUId); ce(2023);

    // get egsphant file
    GetPrivateProfileString("General", "egsphant file", "<<missing>>", charBuffer, CHARLEN, input_file);
	string egsphant_file_input = string(charBuffer);
	if (egsphant_file_input == "<<missing>>")
		error(2011, "Property \"egsphant file\" is not speciefied in the section \"General\" in the input file \"%s\".\n", input_file);
    
    // get the absolute path of the egsphant file
    GetFullPathName(egsphant_file_input.c_str(), CHARLEN, charBuffer, NULL);
    egsphant_file_input = string(charBuffer);
    egsphant_file = egsphant_file_input.c_str();
    logstr("  Phantom . . . . . . . . . . %s\n", egsphant_file);

    // get pegs file
    GetPrivateProfileString("General", "pegs file", "<<missing>>", charBuffer, CHARLEN, input_file);
	string pegs_file_input = string(charBuffer);
	if (pegs_file_input == "<<missing>>")
		error(2012, "Property \"pegs file\" is not speciefied in the section \"General\" in the input file \"%s\".\n", input_file);
    
    // get the absolute path of the pegs file
    GetFullPathName(pegs_file_input.c_str(), CHARLEN, charBuffer, NULL);
    pegs_file_input = string(charBuffer);
    pegs_file = pegs_file_input.c_str();
    logstr("  PEGS4 file  . . . . . . . . %s\n", pegs_file);

    // get histories
    GetPrivateProfileString("General", "histories", "<<missing>>", charBuffer, CHARLEN, input_file);
	if (string(charBuffer) == "<<missing>>")
		error(2013, "Property \"histories\" is not speciefied in the section \"General\" in the input file \"%s\".\n", input_file);
	long long int num_histories_input;
	if (sscanf(charBuffer, "%I64d", &num_histories_input) != 1)
		error(2014, "Could not parse the number of histories \"%s\" in the input file \"%s\" as an 8-byte integer.\n", charBuffer, input_file);
    if (num_histories_input <= 0)
        error(2015, "The number of histories must be positive, but %I64d was specified in the input file \"%s\".\n", num_histories_input, input_file);
    ulong num_histories = (ulong)num_histories_input;
    logstr("  Histories . . . . . . . . . %I64u\n", num_histories);

    // get RNG seed
    bool defaultRNGSeedUsed = false;
    if (!RNGSeedGiven) {
        GetPrivateProfileString("General", "RNG seed", "<<missing>>", charBuffer, CHARLEN, input_file);
	    if (string(charBuffer) == "<<missing>>") {
            defaultRNGSeedUsed = true;
		    time_t t;
            time(&t);
            seed = (uint)t;
        }
        else {
            uint seed_input;
	        if (sscanf(charBuffer, "%u", &seed_input) != 1)
		        error(2016, "Could not parse the RNG seed \"%s\" in the input file \"%s\" as an unsigned 4-byte integer.\n", charBuffer, input_file);
            seed = seed_input;
        }
    }
    if (defaultRNGSeedUsed)
		logstr("WARNING: No RNG seed was specified. A seed was automatically generated based on the current time.\n");
    logstr("  RNG seed  . . . . . . . . . %u\n", seed);

    // get data specifications
    GetPrivateProfileString("Data", "data directory", "<<missing>>", charBuffer, CHARLEN, input_file);
	string data_dir_input = string(charBuffer);
    if (data_dir_input == "<<missing>>")
		error(2017, "Property \"data directory\" is not speciefied in the section \"Data\" in the input file \"%s\".\n", input_file);
    data_dir_input += "\\";

    // get the absolute path of the data directory
    GetFullPathName(data_dir_input.c_str(), CHARLEN, charBuffer, NULL);
    data_dir_input = string(charBuffer);
    data_dir = data_dir_input.c_str();
    logstr("  Data directory  . . . . . . %s\n", data_dir);
	
    GetPrivateProfileString("Data", "MT parameter file", "<<missing>>", charBuffer, CHARLEN, input_file);
	string MT_params_file_input = string(charBuffer);
    if (MT_params_file_input == "<<missing>>")
		error(2018, "Property \"MT parameter file\" is not speciefied in the section \"Data\" in the input file \"%s\".\n", input_file);
    MT_params_file = MT_params_file_input.c_str();
    logstr("  MT parameter file . . . . . %s\n", MT_params_file);

    GetPrivateProfileString("Data", "photon xsections", "<<missing>>", charBuffer, CHARLEN, input_file);
	string photon_xsections_input = string(charBuffer);
    if (photon_xsections_input == "<<missing>>")
		error(2019, "Property \"photon xsections\" is not speciefied in the section \"Data\" in the input file \"%s\".\n", input_file);
    photon_xsections = photon_xsections_input.c_str();
    logstr("  Photon xsections  . . . . . %s\n", photon_xsections);

    GetPrivateProfileString("Data", "atomic ff file", "<<missing>>", charBuffer, CHARLEN, input_file);
	string atomic_ff_file_input = string(charBuffer);
    if (atomic_ff_file_input == "<<missing>>")
		error(2020, "Property \"atomic ff file\" is not speciefied in the section \"Data\" in the input file \"%s\".\n", input_file);
    atomic_ff_file = atomic_ff_file_input.c_str();
    logstr("  Atomic ff file  . . . . . . %s\n", atomic_ff_file);

    // write settings
    logstr("\nSettings\n");
    logstr("  Warps per block . . . . . . %d\n", SIMULATION_WARPS_PER_BLOCK);
    logstr("  Blocks per multiprocessor . %d\n", SIMULATION_BLOCKS_PER_MULTIPROC);
    logstr("  Iterations outer loop . . . %d\n", SIMULATION_ITERATIONS);
#ifdef USE_ENERGY_SPECTRUM
    logstr("  USE_ENERGY_SPECTRUM . . . . enabled\n");
#else
    logstr("  USE_ENERGY_SPECTRUM . . . . disabled\n");
#endif
#ifdef DO_LIST_DEPTH_COUNT
    logstr("  DO_LIST_DEPTH_COUNT . . . . enabled\n");
#else
    logstr("  DO_LIST_DEPTH_COUNT . . . . disabled\n");
#endif
    
    // perform initialization
    init(seed);
    
    clock_t tic2 = clock();

    cudaEvent_t start, stop; 
    cudaEventCreate(&start); ce(11001);
    cudaEventCreate(&stop); ce(11002);

    float elapsed;
    float time_sim = 0.0F;
    float time_sum = 0.0F;
    uint time_copy = 0;

    ulong this_total[NUM_CAT];
    ulong grand_total[NUM_CAT];
    for (uchar i = 0; i < NUM_CAT; i++)
        grand_total[i] = 0;

    bool limit_reached = grand_total[p_new_particle] >= num_histories;
    ulong num_in_progress = 0;
    bool init = true;

    // list depth counter
#ifdef DO_LIST_DEPTH_COUNT
    ulong list_depth = 0;
    ulong num_it = 0;
#endif
    
    // log time
	getTime();
	logstr("\n\nStarting simulation at %s\n\n", charBuffer);
    
    printf("simulation running, wait for ETA...");
    do {
        // do simulation step
        cudaEventRecord(start); ce(11003);
        simulation_step_kernel<<<dim3(SIMULATION_BLOCKS_PER_MULTIPROC, NUM_MULTIPROC), SIMULATION_WARPS_PER_BLOCK * WARP_SIZE>>>(init, limit_reached);
        cudaEventRecord(stop); ce(11004);
        cudaEventSynchronize(stop); ce(11005);
        cudaEventElapsedTime(&elapsed, start, stop); ce(11006);
        time_sim += elapsed;
        init = false;

        // sum detector scores
        cudaEventRecord(start); ce(11007);
        sum_detector_scores_kernel<<<SUM_DETECTOR_NUM_BLOCKS, SUM_DETECTOR_WARPS_PER_BLOCK * WARP_SIZE>>>();
        cudaEventRecord(stop); ce(11008);
        cudaEventSynchronize(stop); ce(11009);
        cudaEventElapsedTime(&elapsed, start, stop); ce(11010);
        time_sum += elapsed;

        // copy counts from device
        time_copy += read_step_counts(this_total, grand_total);
        ulong num_finished_histories = grand_total[p_new_particle];
        limit_reached = num_finished_histories >= num_histories;

        // list depth counter
#ifdef DO_LIST_DEPTH_COUNT
        cudaMemcpy(h_total_list_depth, d_total_list_depth, sizeof(total_list_depth_t), cudaMemcpyDeviceToHost); ce(11011);
        cudaMemcpy(h_total_num_inner_iterations, d_total_num_inner_iterations, sizeof(total_num_inner_iterations_t), cudaMemcpyDeviceToHost); ce(11012);
        for (uchar i = 0; i < SIMULATION_NUM_BLOCKS; i++) {
            list_depth += (*h_total_list_depth)[i];
            num_it += (*h_total_num_inner_iterations)[i];
        }
#endif

        // count number of particles in progress
        num_in_progress = 0;
        num_in_progress += this_total[p_cutoff_discard];
        num_in_progress += this_total[p_user_discard];
        num_in_progress += this_total[p_photon_step];
        num_in_progress += this_total[p_rayleigh];
        num_in_progress += this_total[p_compton];
        num_in_progress += this_total[p_photo];
        num_in_progress += this_total[p_pair];
        num_in_progress += this_total[p_new_particle];

        // calculate ETA and display progress
        clock_t tac = clock();
        elapsed = (float)(tac - tic) / (float)CLOCKS_PER_SEC;
        float complete = (float)num_finished_histories / (float)num_histories;
        float eta = elapsed / complete - elapsed;
        if (eta < 0.0F)
            eta = 0.0F;

        printf("\r%I64u (%.2f%%) histories started, elapsed time: %.0f, ETA: %.0f   ", num_finished_histories, 100.0F * complete, elapsed, eta);

    } while (num_in_progress > 0);
    printf("\r");

    // log time
	getTime();
	logstr("Finished simulation at %s\n\n", charBuffer);

    logstr("\nSimulation step counts\n");
    logstr("  Cutoff discard  . . . . . . %I64u\n", grand_total[p_cutoff_discard]);
    logstr("  User discard  . . . . . . . %I64u\n", grand_total[p_user_discard]);
    logstr("  Photon step . . . . . . . . %I64u\n", grand_total[p_photon_step]);
    logstr("  Rayleigh  . . . . . . . . . %I64u\n", grand_total[p_rayleigh]);
    logstr("  Compton . . . . . . . . . . %I64u\n", grand_total[p_compton]);
    logstr("  Photo . . . . . . . . . . . %I64u\n", grand_total[p_photo]);
    logstr("  Pair  . . . . . . . . . . . %I64u\n", grand_total[p_pair]);
    logstr("  New particles . . . . . . . %I64u\n", grand_total[p_new_particle]);

    // list depth counter
#ifdef DO_LIST_DEPTH_COUNT
    logstr("\nDivergence\n");
    logstr("  Total different steps . . . %I64u\n", list_depth);
    logstr("  Total iterations  . . . . . %I64u\n", num_it);
    logstr("  Average different steps . . %f\n", (double)list_depth / (double)num_it);
    logstr("  Average active threads  . . %.2f %% (at least)\n", 100.0 * (double)num_it / (double)list_depth);
#endif

    cudaEventDestroy(start); ce(11013);
    cudaEventDestroy(stop); ce(11014);

    // copy results
    clock_t copy_tic = clock();
    uint size_det = h_detector.N.x * h_detector.N.y * sizeof(double);

    double *detector_results_count[NUM_DETECTOR_CAT + 1];
    double *detector_results_energy[NUM_DETECTOR_CAT + 1];
    detector_results_count[NUM_DETECTOR_CAT] = (double*)malloc(size_det);
    detector_results_energy[NUM_DETECTOR_CAT] = (double*)malloc(size_det);

    for (uchar i = 0; i < NUM_DETECTOR_CAT; i++) {
        detector_results_count[i] = (double*)malloc(size_det);
        detector_results_energy[i] = (double*)malloc(size_det);

        // copy detector totals from device
        cudaMemcpy(detector_results_count[i], d_detector_totals_count[i], size_det, cudaMemcpyDeviceToHost); ce(11015, "i = %d", i);
        cudaMemcpy(detector_results_energy[i], d_detector_totals_energy[i], size_det, cudaMemcpyDeviceToHost); ce(11016, "i = %d", i);
    }

    total_weights_t h_total_weights;
    cudaMemcpy(h_total_weights, d_total_weights, sizeof(total_weights_t), cudaMemcpyDeviceToHost); ce(11017);

    clock_t copy_tac = clock();
    time_copy += (copy_tac - copy_tic);

    double total_weight = 0.0F;
    for (uchar i = 0; i < SIMULATION_NUM_BLOCKS; i++)
        total_weight += h_total_weights[i];

    double average_weight = total_weight / (double)grand_total[p_new_particle];
    for (uint i = 0; i < h_detector.N.x * h_detector.N.y; i++) {
        double total_count = 0.0F;
        double total_energy = 0.0F;

        for (uchar j = 0; j < NUM_DETECTOR_CAT; j++) {
            detector_results_count[j][i] /= average_weight;
            detector_results_energy[j][i] /= average_weight;

            total_count += detector_results_count[j][i];
            total_energy += detector_results_energy[j][i];
        }

        detector_results_count[NUM_DETECTOR_CAT][i] = total_count;
        detector_results_energy[NUM_DETECTOR_CAT][i] = total_energy;
    }

    write_output(rootName, "count", detector_results_count);
    write_output(rootName, "energy", detector_results_energy);

    for (uchar i = 0; i <= NUM_DETECTOR_CAT; i++) {
        free(detector_results_count[i]);
        free(detector_results_energy[i]);
    }

    free_all();
    
    clock_t tac = clock();

    float time_copy_f = (float)time_copy / (float)CLOCKS_PER_SEC * 1000.0F;
    float total_time = (float)(tac - tic) / (float)CLOCKS_PER_SEC * 1000.0F;
    float init_time = (float)(tic2 - tic) / (float)CLOCKS_PER_SEC * 1000.0F;
    float total_cpu_time = (float)(tac - tic2) / (float)CLOCKS_PER_SEC * 1000.0F;
    float other = total_time - time_copy_f - time_sim - time_sum;

    logstr("\nTiming statistics\n", total_time);
    logstr("  Elapsed time  . . . . . . . %.2f ms (%.2f %%)\n", total_time, 100.0);
    logstr("  Total CPU/GPU . . . . . . . %.2f ms (%.2f %%)\n", total_cpu_time, 100.0F * total_cpu_time / total_time);
    logstr("  Simulation kernel . . . . . %.2f ms (%.2f %%)\n", time_sim, 100.0F * time_sim / total_time);
    logstr("  Summing kernel  . . . . . . %.2f ms (%.2f %%)\n", time_sum, 100.0F * time_sum / total_time);
    logstr("  Copying . . . . . . . . . . %.2f ms (%.2f %%)\n", time_copy_f, 100.0F * time_copy_f / total_time);
    logstr("  Other . . . . . . . . . . . %.2f ms (%.2f %%)\n", other, 100.0F * other / total_time);
    logstr("  Initialization. . . . . . . %.2f ms (%.2f %%)\n", init_time, 100.0F * init_time / total_time);
    
    logstr("\nHistories per ms: %f\n", (float)grand_total[p_new_particle] / total_time);
    
    // log time
	getTime();
	logstr("\n\nCUDA EGS finished at %s\n", charBuffer);

    fclose(logfile);

    return 0;
}

